@@ -2,24 +2,31 @@ import sys
 import os
 import re
 import random
-import numpy as np
 from itertools import chain, count
 import pickle
+import gzip
 import argparse
 import importlib.resources
+from contextlib import closing
 from apycula import codegen
 from apycula import chipdb
 from apycula import attrids
-from apycula.bslib import read_bitstream
-from apycula.wirenames import wirenames
+from apycula.bslib import read_bitstream, display
 
 _device = ""
 _pinout = ""
 _packages = {
         'GW1N-1' : 'LQFP144', 'GW1NZ-1' : 'QFN48', 'GW1N-4' : 'PBGA256', 'GW1N-9C' : 'UBGA332',
-        'GW1N-9' : 'PBGA256', 'GW1NS-4' : 'QFN48', 'GW1NS-2' : 'LQFP144',
+        'GW1N-9' : 'PBGA256', 'GW1NS-4' : 'QFN48P', 'GW1NS-2' : 'LQFP144', 'GW2A-18': 'PBGA256',
+        'GW2A-18C' : 'PBGA256S', 'GW5A-25A' : 'MBGA121N'
 }
 
+def print_sorted_dict(start, d):
+    print(start, end='{')
+    for i in sorted(d):
+        print(f'{i}:{d[i]}, ', end='')
+    print('}')
+
 # bank iostandards
 # XXX default io standard may be board-dependent!
 _banks = {'0': "LVCMOS18", '1': "LVCMOS18", '2': "LVCMOS18", '3': "LVCMOS18"}
@@ -53,10 +60,11 @@ def _io_mode_sort_func(mode):
     return l
 
 #
-def get_attr_name(attrname_table, code):
+def get_attr_name(attrname_table, code, tableName):
     for name, cod in attrname_table.items():
         if cod == code:
             return name
+    print(f'Unknown attr name for table: {tableName} code:{code}')
     return ''
 
 # fix names and types of the PLL attributes
@@ -182,7 +190,7 @@ def get_dff_type(dff_idx, in_attrs):
 # parse attributes and values use 'logicinfo' table
 # returns {attr: value}
 # attribute names are decoded with the attribute table, but the values are returned in raw form
-def parse_attrvals(tile, logicinfo_table, fuse_table, attrname_table):
+def parse_attrvals(tile, logicinfo_table, fuse_table, attrname_table, tableName):
     def is_neg_key(key):
         for k in key:
             if k < 0:
@@ -207,7 +215,7 @@ def parse_attrvals(tile, logicinfo_table, fuse_table, attrname_table):
             zero_mask.update(bits)
         else:
             set_mask.update(bits)
-    set_bits =  {(row, col) for row, col in set_mask if tile[row][col] == 1}
+    set_bits = {(row, col) for row, col in set_mask if tile[row][col] == 1}
     neg_bits = {(row, col) for row, col in zero_mask if tile[row][col] == 1}
 
     # find candidates from fuse table
@@ -225,7 +233,7 @@ def parse_attrvals(tile, logicinfo_table, fuse_table, attrname_table):
             attrvals.update(clean_av) # set attributes
             for idx in clean_av:
                 attr, val = logicinfo_table[idx]
-                res[get_attr_name(attrname_table, attr)] = val
+                res[get_attr_name(attrname_table, attr, tableName)] = val
 
     # records with a negative keys and used fuses
     neg_attrvals = set()
@@ -238,7 +246,7 @@ def parse_attrvals(tile, logicinfo_table, fuse_table, attrname_table):
                 keep = False
                 break
             for idx in av:
-                attr, _ = logicinfo_table[idx]
+                attr, _ = logicinfo_table[abs(idx)]
                 if attr in res.keys():
                     keep = False
                     break
@@ -248,7 +256,7 @@ def parse_attrvals(tile, logicinfo_table, fuse_table, attrname_table):
 
     for idx in neg_attrvals:
         attr, val = logicinfo_table[idx]
-        res[get_attr_name(attrname_table, attr)] = val
+        res[get_attr_name(attrname_table, attr, tableName)] = val
 
     # records with a negative keys and unused fuses
     cnd = {av for av, bits in fuse_table.items() if is_neg_key(av) and not bits.issubset(neg_bits)}
@@ -261,7 +269,7 @@ def parse_attrvals(tile, logicinfo_table, fuse_table, attrname_table):
         if keep:
             for idx in get_negative(av):
                 attr, val = logicinfo_table[idx]
-                res[get_attr_name(attrname_table, attr)] = val
+                res[get_attr_name(attrname_table, attr, tableName)] = val
     return res
 
 # { (row, col, type) : idx}
@@ -281,20 +289,38 @@ def get_pll_A(db, row, col, typ):
     return row, col, 'A'
 
 _iologic_mode = {
-        'MODDRX2': 'OSER4',  'ODDRX2': 'OSER4',
-        'MODDRX4': 'OSER8',  'ODDRX4': 'OSER8',
-        'MODDRX5': 'OSER10', 'ODDRX5': 'OSER10',
-        'VIDEORX': 'OVIDEO', 'ODDRX8': 'OSER16',
-        'MIDDRX2': 'IDES4',  'IDDRX2': 'IDES4',
-        'MIDDRX4': 'IDES8',  'IDDRX4': 'IDES8',
-        'MIDDRX5': 'IDES10', 'IDDRX5': 'IDES10',
-        'IDDRX8': 'IDES16',
+        'MODDRX2':  'OSER4',  'ODDRX2': 'OSER4',
+        'MODDRX21': 'OSER4',  'ODDRX2': 'OSER4',
+        'MODDRX4':  'OSER8',  'ODDRX4': 'OSER8',
+        'MODDRX5':  'OSER10', 'ODDRX5': 'OSER10',
+        'VIDEORX':  'OVIDEO', 'ODDRX8': 'OSER16',
+        'MIDDRX2':  'IDES4',  'IDDRX2': 'IDES4',
+        'MIDDRX4':  'IDES8',  'IDDRX4': 'IDES8',
+        'MIDDRX5':  'IDES10', 'IDDRX5': 'IDES10',
+        'IDDRX8':   'IDES16',
         }
+
+# BSRAM has 3 cells: BSRAM, BSRAM0 and BSRAM1
+# { (row, col) : idx }
+_bsram_cells = {}
+def get_bsram_main_cell(db, row, col, typ):
+    if typ[-4:] == '_AUX':
+        col -= 1
+        if 'BSRAM_AUX' in db.grid[row][col].bels:
+            col -= 1
+    return row, col
+
+# The DSP has 9 cells: the main one and a group of auxiliary ones.
+def get_dsp_main_cell(db, row, col, typ):
+    if typ[-6:-2] == '_AUX':
+        col = 1 + (col - 1) // 9
+    return row, col
+
 # noiostd --- this is the case when the function is called
 # with iostd by default, e.g. from the clock fuzzer
 # With normal gowin_unpack io standard is determined first and it is known.
 # (bels, pips, clock_pips)
-def parse_tile_(db, row, col, tile, default=True, noalias=False, noiostd = True):
+def parse_tile_(db, row, col, tile, bm=None, default=True, noiostd = True):
     if not _bank_fuse_tables:
         # create bank fuse table
         for ttyp in db.longval.keys():
@@ -306,14 +332,32 @@ def parse_tile_(db, row, col, tile, default=True, noalias=False, noiostd = True)
     skip_bels = set()
     #print((row, col))
     tiledata = db.grid[row][col]
+    #if 'HCLK' in db.shortval[tiledata.ttyp].keys():
+    #    attrvals =parse_attrvals(tile, db.logicinfo['HCLK'], db.shortval[tiledata.ttyp]['HCLK'], attrids.hclk_attrids)
+    #    if attrvals:
+    #        print(row, col, attrvals)
+
+    #if tiledata.ttyp in db.longfuses:
+    #    if 'DLLDEL0' in db.longfuses[tiledata.ttyp].keys():
+    #        attrvals =parse_attrvals(tile, db.logicinfo['DLLDLY'], db.longfuses[tiledata.ttyp]['DLLDEL0'], attrids.dlldly_attrids)
+    #        if attrvals:
+    #            print(row, col, attrvals)
+    #    if 'DLLDEL1' in db.longfuses[tiledata.ttyp].keys():
+    #        attrvals =parse_attrvals(tile, db.logicinfo['DLLDLY'], db.longfuses[tiledata.ttyp]['DLLDEL1'], attrids.dlldly_attrids)
+    #        if attrvals:
+    #            print(row, col, attrvals)
+
     clock_pips = {}
     bels = {}
     for name, bel in tiledata.bels.items():
+        if name.startswith("ADC"):
+            attrvals = parse_attrvals(tile, db.rev_logicinfo('ADC'), db.shortval[tiledata.ttyp]['ADC'], attrids.adc_attrids, "ADC")
+            print(row, col, name, tiledata.ttyp, attrvals)
         if name.startswith("RPLL"):
             idx = _pll_cells.setdefault(get_pll_A(db, row, col, name[4]), len(_pll_cells))
             modes = { f'DEVICE="{_device}"' }
             if 'PLL' in db.shortval[tiledata.ttyp].keys():
-                attrvals = pll_attrs_refine(parse_attrvals(tile, db.logicinfo['PLL'], db.shortval[tiledata.ttyp]['PLL'], attrids.pll_attrids))
+                attrvals = pll_attrs_refine(parse_attrvals(tile, db.rev_logicinfo('PLL'), db.shortval[tiledata.ttyp]['PLL'], attrids.pll_attrids, "PLL"))
                 for attrval in attrvals:
                     modes.add(attrval)
             if modes:
@@ -321,7 +365,7 @@ def parse_tile_(db, row, col, tile, default=True, noalias=False, noiostd = True)
             continue
         if name == "PLLVR":
             idx = _pll_cells.setdefault(get_pll_A(db, row, col, 'A'), len(_pll_cells))
-            attrvals = pll_attrs_refine(parse_attrvals(tile, db.logicinfo['PLL'], db.shortval[tiledata.ttyp]['PLL'], attrids.pll_attrids))
+            attrvals = pll_attrs_refine(parse_attrvals(tile, db.rev_logicinfo('PLL'), db.shortval[tiledata.ttyp]['PLL'], attrids.pll_attrids, "PLL"))
             modes = { f'DEVICE="{_device}"' }
             for attrval in attrvals:
                 modes.add(attrval)
@@ -329,19 +373,50 @@ def parse_tile_(db, row, col, tile, default=True, noalias=False, noiostd = True)
                 bels[f'{name}{idx}'] = modes
             continue
         if name.startswith("OSC"):
-            attrvals = osc_attrs_refine(parse_attrvals(tile, db.logicinfo['OSC'], db.shortval[tiledata.ttyp]['OSC'], attrids.osc_attrids))
+            attrvals = osc_attrs_refine(parse_attrvals(tile, db.rev_logicinfo('OSC'), db.shortval[tiledata.ttyp]['OSC'], attrids.osc_attrids, "OSC"))
             modes = set()
             for attrval in attrvals:
                 modes.add(attrval)
             if modes:
                 bels[name] = modes
             continue
-        if name.startswith("IOLOGIC"):
-            idx = name[-1]
-            attrvals = parse_attrvals(tile, db.logicinfo['IOLOGIC'], db.shortval[tiledata.ttyp][f'IOLOGIC{idx}'], attrids.iologic_attrids)
-            #print(row, col, attrvals)
+        if name.startswith("BSRAM"):
+            # disabled BSRAM cells have no fuse tables
+            if 'BSRAM_SP' not in db.shortval[tiledata.ttyp]:
+                continue
+            idx = _bsram_cells.setdefault(get_bsram_main_cell(db, row, col, name), len(_bsram_cells))
+            #print(row, col, name, idx, tiledata.ttyp)
+            attrvals = parse_attrvals(tile, db.rev_logicinfo('BSRAM'), db.shortval[tiledata.ttyp]['BSRAM_SP'], attrids.bsram_attrids, "BSRAM")
             if not attrvals:
                 continue
+            #print(row, col, name, idx, tiledata.ttyp, attrvals)
+            bels[f'{name}{idx}'] = {}
+            continue
+        if name.startswith("ALU54D"):
+            continue
+        if name.startswith("DSP") or name.startswith("DSP_AUX"):
+            modes = set()
+            idx = name[-1]
+            #print(row, col, name, idx, tiledata.ttyp)
+            if name.startswith("DSP_AUX"):
+                row, col = get_dsp_main_cell(db, row, col, name)
+
+            if f'DSP{idx}' in db.shortval[tiledata.ttyp]:
+                attrvals = parse_attrvals(tile, db.rev_logicinfo('DSP'), db.shortval[tiledata.ttyp][f'DSP{idx}'], attrids.dsp_attrids, "DSP")
+                #print_sorted_dict(f'{row}, {col}, {name}, {idx}, {tiledata.ttyp} - ', attrvals)
+                for attrval in attrvals:
+                    modes.add(attrval)
+            if modes and not name.startswith("DSP_AUX"):
+                bels[f'{name}{idx}'] = modes
+            continue
+        if name.startswith("IOLOGIC"):
+            idx = name[-1]
+            attrvals = parse_attrvals(tile, db.rev_logicinfo('IOLOGIC'), db.shortval[tiledata.ttyp][f'IOLOGIC{idx}'], attrids.iologic_attrids, "IOLOGIC")
+            if not attrvals:
+                continue
+            # additional IOLOGIC components
+            # XXX delays and FFs in IO
+            # main component
             if 'OUTMODE' in attrvals.keys():
                 # XXX skip oddr
                 if attrvals['OUTMODE'] in {attrids.iologic_attrvals['MODDRX1'], attrids.iologic_attrvals['ODDRX1']}:
@@ -376,51 +451,91 @@ def parse_tile_(db, row, col, tile, default=True, noalias=False, noiostd = True)
                 bels.setdefault(name, set()).add(f"CLKODDRMUX_ECLK={attrids.iologic_num2val[attrvals['CLKODDRMUX_ECLK']]}")
         if name.startswith("DFF"):
             idx = int(name[3])
-            attrvals = parse_attrvals(tile, db.logicinfo['SLICE'], db.shortval[tiledata.ttyp][f'CLS{idx // 2}'], attrids.cls_attrids)
-            #print(row, col, attrvals)
+            attrvals = parse_attrvals(tile, db.rev_logicinfo('SLICE'), db.shortval[tiledata.ttyp][f'CLS{idx // 2}'], attrids.cls_attrids, "CLS")
+            #print('parse', row, col, attrvals)
             # skip ALU and unsupported modes
             if attrvals.get('MODE') == attrids.cls_attrvals['SSRAM']:
                 continue
             dff_type = get_dff_type(idx, attrvals)
             if dff_type:
                 bels[f'{name}'] = {dff_type}
+            if f'REG{idx % 2}_SD' in attrvals:
+                bels[f'{name}'].update({'SD'})
             continue
         if name.startswith("IOB"):
             idx = name[-1]
-            attrvals = parse_attrvals(tile, db.logicinfo['IOB'], db.longval[tiledata.ttyp][f'IOB{idx}'], attrids.iob_attrids)
+            io_row, io_col = row, col
+            io_tile = tile
+            io_ttyp = tiledata.ttyp
+
+            if idx == 'B' and _device == 'GW5A-25A' and tiledata.bels[name].fuse_cell_offset:
+                io_row += tiledata.bels[name].fuse_cell_offset[0]
+                io_col += tiledata.bels[name].fuse_cell_offset[1]
+                io_tiledata = db.grid[io_row][io_col]
+                io_tile = bm[io_row, io_col]
+                io_ttyp = io_tiledata.ttyp
+            # XXX
+            if idx == 'B' and 'IOBB' not in db.longval[io_ttyp]:
+                continue
+            attrvals = parse_attrvals(io_tile, db.rev_logicinfo('IOB'), db.longval[io_ttyp][f'IOB{idx}'], attrids.iob_attrids, "IOB")
+            #print(name, io_row, io_col, attrvals)
             try: # we can ask for invalid pin here because the IOBs share some stuff
-                bank = chipdb.loc2bank(db, row, col)
+                bank = chipdb.loc2bank(db, io_row, io_col)
             except KeyError:
                 bank = None
             if attrvals:
                 mode = 'IBUF'
                 if attrvals.get('PERSISTENT', None) == attrids.iob_attrvals['OFF']:
                     mode = 'IOBUF'
-                elif 'ODMUX' in attrvals.keys() or 'ODMUX_1' in attrvals.keys():
+                elif 'ODMUX' in attrvals or 'ODMUX_1' in attrvals:
                     mode = 'OBUF'
                 # Z-1 row 6
                 if _device in {'GW1NZ-1', 'GW1N-1'} and row == 5:
                     mode = 'IOBUF'
-                if 'LVDS_OUT' in attrvals.keys():
+                if 'LVDS_OUT' in attrvals:
                     if mode == 'IOBUF':
                         mode = 'TBUF'
-                    mode = f'TLVDS_{mode}'
+                    if 'MIPI' in attrvals:
+                        mode = 'MIPI_OBUF'
+                    else:
+                        mode = f'TLVDS_{mode}'
                     # skip B bel
                     skip_bels.update({name[:-1] + 'B'})
-                elif idx == 'B' and 'DRIVE' not in attrvals.keys() and 'IO_TYPE' in attrvals.keys():
+                elif 'OD' in attrvals:
+                    mode = 'I3C_IOBUF'
+                elif idx == 'B' and 'DRIVE' not in attrvals and 'IO_TYPE' in attrvals:
                     mode = f'ELVDS_{mode}'
                     # skip B bel
                     skip_bels.update({name})
-                elif 'IOBUF_MIPI_LP' in attrvals.keys():
+                elif 'LPRX_A1' in attrvals:
+                    mode = f'MIPI_IBUF'
+                    # skip B bel
+                    skip_bels.update({name[:-1] + 'B'})
+                elif 'IOBUF_MIPI_LP' in attrvals:
                     mode = f'ELVDS_{mode}'
                     # skip B bel
                     skip_bels.update({name[:-1] + 'B'})
 
                 bels.setdefault(name, set()).add(mode)
         if name.startswith("BANK"):
-            attrvals = parse_attrvals(tile, db.logicinfo['IOB'], _bank_fuse_tables[tiledata.ttyp][name], attrids.iob_attrids)
+            attrvals = parse_attrvals(tile, db.rev_logicinfo('IOB'), _bank_fuse_tables[tiledata.ttyp][name], attrids.iob_attrids, "IOB")
+            #print(name, row, col, attrvals)
+
             for a, v in attrvals.items():
-                bels.setdefault(name, set()).add(f'{a}={attrids.iob_num2val[v]}')
+                bels.setdefault(name, set()).add(f'{a}={attrids.iob_num2val.get(v, str(v))}')
+        if name.startswith("ALU"):
+            idx = int(name[3])
+            attrvals = parse_attrvals(tile, db.rev_logicinfo('SLICE'), db.shortval[tiledata.ttyp][f'CLS{idx // 2}'], attrids.cls_attrids, "CLS")
+            # skip ALU and unsupported modes
+            if attrvals.get('MODE') != attrids.cls_attrvals['ALU']:
+                continue
+            bels[name] = {"C2L"}
+            mode_bits = {(row, col)
+                         for row, col in bel.mode_bits
+                         if tile[row][col] == 1}
+            for mode, bits in bel.modes.items():
+                if bits == mode_bits and (default or bits):
+                    bels[name] = {mode}
         else:
             mode_bits = {(row, col)
                          for row, col in bel.mode_bits
@@ -467,16 +582,18 @@ def parse_tile_(db, row, col, tile, default=True, noalias=False, noiostd = True)
                      for row, col in pip_bits
                      if tile[row][col] == 1}
         for src, bits in srcs.items():
-            # only report connection aliased to by a spine
-            # HCLKs are also switched here, so for now we are also considering SPINExx type wires
-            if bits == used_bits and (noalias or (row, col, src) in db.aliases or (src.startswith('SPINE') and dest.startswith('SPINE'))):
+            if bits == used_bits:
                 clock_pips[dest] = src
 
     # elvds IO uses the B bel bits
     for name in skip_bels:
-        bel_a = bels[f'{name[:-1]}A']
+        bel_a_name = f'{name[:-1]}A'
+        if bel_a_name not in bels:
+            continue
+
+        bel_a = bels[bel_a_name]
         if not bel_a.intersection({'ELVDS_IBUF', 'ELVDS_OBUF', 'ELVDS_IOBUF', 'ELVDS_TBUF',
-                                   'TLVDS_IBUF', 'TLVDS_OBUF', 'TLVDS_IOBUF', 'TLVDS_TBUF'}):
+                                   'TLVDS_IBUF', 'TLVDS_OBUF', 'TLVDS_IOBUF', 'TLVDS_TBUF', 'MIPI_IBUF', 'MIPI_OBUF'}):
             mode = bels[name].intersection({'ELVDS_IBUF', 'ELVDS_OBUF', 'ELVDS_IOBUF', 'ELVDS_TBUF'})
             if mode:
                 old_mode = bel_a.intersection({'IBUF', 'OBUF', 'IOBUF', 'TBUF'})
@@ -509,6 +626,9 @@ iobmap = {
     "ELVDS_TBUF": {"wires": ["I", "OE"], "outputs": ["O", "OB"]},
     "ELVDS_IBUF": {"wires": ["O"], "inputs": ["I", "IB"]},
     "ELVDS_IOBUF": {"wires": ["I", "O", "OE"], "inouts": ["IO", "IOB"]},
+    "MIPI_IBUF": {"wires": ["I", "IB", "OEN", "OENB", "HSREN", "OH", "OL", "OB"], "inouts": ["IO", "IOB"]},
+    "MIPI_OBUF": {"wires": ["I", "IB", "IL", "MODESEL"], "inouts": ["O", "OB"]},
+    "I3C_IOBUF": {"wires": ["I", "O", "MODESEL"], "inouts": ["IO"]},
 }
 
 # OE -> OEN
@@ -603,39 +723,48 @@ def move_iologic(bels):
     return res
 
 def disable_unused_pll_ports(pll):
-    if 'DYN_DA_EN' not in pll.params.keys():
+    if 'DYN_DA_EN' not in pll.params:
         for n in range(0, 4):
-            del pll.portmap[f'PSDA{n}']
-            del pll.portmap[f'DUTYDA{n}']
-            del pll.portmap[f'FDLY{n}']
-    if 'DYN_IDIV_SEL' not in pll.params.keys():
+            if f'PSDA{n}' in pll.portmap:
+                del pll.portmap[f'PSDA{n}']
+                del pll.portmap[f'DUTYDA{n}']
+                del pll.portmap[f'FDLY{n}']
+    if 'DYN_IDIV_SEL' not in pll.params:
         for n in range(0, 6):
-            del pll.portmap[f'IDSEL{n}']
-    if 'DYN_FBDIV_SEL' not in pll.params.keys():
+            if f'IDSEL{n}' in pll.portmap:
+                del pll.portmap[f'IDSEL{n}']
+    if 'DYN_FBDIV_SEL' not in pll.params:
         for n in range(0, 6):
-            del pll.portmap[f'FBDSEL{n}']
-    if 'DYN_ODIV_SEL' not in pll.params.keys():
+            if f'FBDSEL{n}' in pll.portmap:
+                del pll.portmap[f'FBDSEL{n}']
+    if 'DYN_ODIV_SEL' not in pll.params:
         for n in range(0, 6):
-            del pll.portmap[f'ODSEL{n}']
-    if 'PWDEN' in pll.params.keys():
+            if f'ODSEL{n}' in pll.portmap:
+                del pll.portmap[f'ODSEL{n}']
+    if 'PWDEN' in pll.params:
         if pll.params['PWDEN'] == 'DISABLE':
-            del pll.portmap['RESET_P']
+            if 'RESET_P' in pll.portmap:
+                del pll.portmap['RESET_P']
         del pll.params['PWDEN']
-    if 'RSTEN' in pll.params.keys():
+    if 'RSTEN' in pll.params:
         if pll.params['RSTEN'] == 'DISABLE':
-            del pll.portmap['RESET']
+            if 'RESET' in pll.portmap:
+                del pll.portmap['RESET']
         del pll.params['RSTEN']
-    if 'CLKOUTDIV3' in pll.params.keys():
+    if 'CLKOUTDIV3' in pll.params:
         if pll.params['CLKOUTDIV3'] == 'DISABLE':
-            del pll.portmap['CLKOUTD3']
+            if 'CLKOUTD3' in pll.portmap:
+                del pll.portmap['CLKOUTD3']
         del pll.params['CLKOUTDIV3']
-    if 'CLKOUTDIV' in pll.params.keys():
+    if 'CLKOUTDIV' in pll.params:
         if pll.params['CLKOUTDIV'] == 'DISABLE':
-            del pll.portmap['CLKOUTD']
+            if 'CLKOUTD' in pll.portmap:
+                del pll.portmap['CLKOUTD']
         del pll.params['CLKOUTDIV']
-    if 'CLKOUTPS' in pll.params.keys():
+    if 'CLKOUTPS' in pll.params:
         if pll.params['CLKOUTPS'] == 'DISABLE':
-            del pll.portmap['CLKOUTP']
+            if 'CLKOUTP' in pll.portmap:
+                del pll.portmap['CLKOUTP']
         del pll.params['CLKOUTPS']
 
 _tbrlre = re.compile(r"IO([TBRL])(\d+)(\w)")
@@ -671,14 +800,16 @@ def modify_pll_inputs(db, pll):
             if insel == 'CLKIN0':
                 find_pll_in_pin(db, pll)
             else:
-                del pll.portmap['CLKIN']
+                if 'CLKIN' in pll.portmap:
+                    del pll.portmap['CLKIN']
         del pll.params['INSEL']
     if 'FBSEL' in pll.params.keys():
         fbsel = pll.params['FBSEL']
         if fbsel == 'CLKFB3':
             # internal
             pll.params['CLKFB_SEL'] = '"internal"'
-            del pll.portmap['CLKFB']
+            if 'CLKFB' in pll.portmap:
+                del pll.portmap['CLKFB']
         elif fbsel == 'CLKFB0':
             # external CLK2
             pll.params['CLKFB_SEL'] = '"external"'
@@ -746,14 +877,14 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
     # db is 0-based, floorplanner is 1-based
     row = dbrow+1
     col = dbcol+1
-    aliases = db.grid[dbrow][dbcol].aliases
-    for dest, src in chain(pips.items(), aliases.items(), clock_pips.items()):
+
+    for dest, src in chain(pips.items(), clock_pips.items()):
         srcg = chipdb.wire2global(row, col, db, src)
         destg = chipdb.wire2global(row, col, db, dest)
         mod.wires.update({srcg, destg})
         mod.assigns.append((destg, srcg))
 
-    belre = re.compile(r"(IOB|LUT|DFF|BANK|CFG|ALU|RAM16|ODDR|OSC[ZFHWO]?|BUFS|RPLL[AB]|PLLVR|IOLOGIC)(\w*)")
+    belre = re.compile(r"(IOB|LUT|DFF|BANK|CFG|ALU|RAM16|ODDR|OSC[ZFHWO]?|BUFS|RPLL[AB]|PLLVR|IOLOGIC|BSRAM|DSP)(\w*)")
     bels_items = move_iologic(bels)
 
     iologic_detected = set()
@@ -850,7 +981,7 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
             for paramval in flags:
                 param, _, val = paramval.partition('=')
                 pll.params[param] = val
-            portmap = db.grid[dbrow][dbcol].bels[bel[:-1]].portmap
+            portmap = db.grid[dbrow][dbcol].bels[bel[:5]].portmap
             for port, wname in portmap.items():
                 pll.portmap[port] = f"R{row}C{col}_{wname}"
         elif typ.startswith("PLLVR"):
@@ -862,6 +993,19 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
             portmap = db.grid[dbrow][dbcol].bels[bel[:-1]].portmap
             for port, wname in portmap.items():
                 pll.portmap[port] = f"R{row}C{col}_{wname}"
+        elif typ.startswith("BSRAM"):
+            #print(dbrow, dbcol, typ, bel, idx)
+            if idx.startswith("_AUX"):
+                continue
+            bel_name = "BSRAM"
+            name = f"BSRAM_{idx}"
+            pll = mod.primitives.setdefault(name, codegen.Primitive("BSRAM", name))
+            for paramval in flags:
+                param, _, val = paramval.partition('=')
+                pll.params[param] = val
+            portmap = db.grid[dbrow][dbcol].bels[bel_name].portmap
+            for port, wname in portmap.items():
+                pll.portmap[port] = f"R{row}C{col}_{wname}"
         elif typ == "ALU":
             #print(flags)
             kind, = flags # ALU only have one flag
@@ -869,11 +1013,13 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
             name = f"R{row}C{col}_ALU_{idx}"
             if kind == 'hadder':
                 kind = '0'
-            if kind in "012346789": # main ALU
+            if kind in "012346789" or kind == "C2L" : # main ALU
                 alu = codegen.Primitive("ALU", name)
                 alu.params["ALU_MODE"] = kind
-                alu.portmap['SUM'] = f"R{row}C{col}_F{idx}"
+                if kind != "C2L":
+                    alu.portmap['SUM'] = f"R{row}C{col}_F{idx}"
                 alu.portmap['CIN'] = f"R{row}C{col}_CIN{idx}"
+                alu.portmap['I2'] = f"R{row}C{col}_C{idx}"
                 if idx != 5:
                     alu.portmap['COUT'] = f"R{row}C{col}_CIN{idx+1}"
                 else:
@@ -886,6 +1032,12 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
                 elif kind == "0":
                     alu.portmap['I0'] = f"R{row}C{col}_B{idx}"
                     alu.portmap['I1'] = f"R{row}C{col}_D{idx}"
+                    alu.portmap['I3'] = f"R{row}C{col}_A{idx}"
+                elif kind == "C2L":
+                    alu.portmap['I0'] = f"R{row}C{col}_B{idx}"
+                    alu.portmap['I1'] = f"R{row}C{col}_D{idx}"
+                    alu.portmap['COUT'] = f"R{row}C{col}_F{idx}"
+                    alu.params["ALU_MODE"] = "9" # XXX
                 else:
                     alu.portmap['I0'] = f"R{row}C{col}_A{idx}"
                     alu.portmap['I1'] = f"R{row}C{col}_D{idx}"
@@ -902,12 +1054,12 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
             ram16.params["INIT_1"] = f"16'b{val1:016b}"
             ram16.params["INIT_2"] = f"16'b{val2:016b}"
             ram16.params["INIT_3"] = f"16'b{val3:016b}"
-            ram16.portmap['DI'] = [f"R{row}C{col}_{x}5" for x in "ABCD"]
+            ram16.portmap['DI'] = [f"R{row}C{col}_{x}5" for x in "DCBA"]
             ram16.portmap['CLK'] = f"R{row}C{col}_CLK2"
             ram16.portmap['WRE'] = f"R{row}C{col}_LSR2"
-            ram16.portmap['WAD'] = [f"R{row}C{col}_{x}4" for x in "ABCD"]
-            ram16.portmap['RAD'] = [f"R{row}C{col}_{x}0" for x in "ABCD"]
-            ram16.portmap['DO'] = [f"R{row}C{col}_F{x}" for x in range(4)]
+            ram16.portmap['WAD'] = [f"R{row}C{col}_{x}4" for x in "DCBA"]
+            ram16.portmap['RAD'] = [f"R{row}C{col}_{x}0" for x in "DCBA"]
+            ram16.portmap['DO'] = [f"R{row}C{col}_F{x}" for x in range(4, -1, -1)]
             mod.wires.update(chain.from_iterable([x if isinstance(x, list) else [x] for x in ram16.portmap.values()]))
             mod.primitives[name] = ram16
         elif typ in {"OSC", "OSCZ", "OSCF", "OSCH", "OSCW", "OSCO"}:
@@ -923,6 +1075,10 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
             mod.primitives[name] = osc
         elif typ == "DFF":
             #print(flags)
+            sd = False
+            if 'SD' in flags:
+                sd = True
+                flags.remove('SD')
             kind, = flags # DFF only have one flag
             if kind == "RAM": continue
             idx = int(idx)
@@ -930,7 +1086,10 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
             name = f"R{row}C{col}_{typ}E_{idx}"
             dff = codegen.Primitive(kind+"E", name)
             dff.portmap['CLK'] = f"R{row}C{col}_CLK{idx//2}"
-            dff.portmap['D'] = f"R{row}C{col}_F{idx}"
+            if sd:
+                dff.portmap['D'] = f"R{row}C{col}_SEL{idx}"
+            else:
+                dff.portmap['D'] = f"R{row}C{col}_F{idx}"
             dff.portmap['Q'] = f"R{row}C{col}_Q{idx}"
             dff.portmap['CE'] = f"R{row}C{col}_CE{idx//2}"
             if port:
@@ -944,7 +1103,28 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
             except ValueError:
                 continue
             flags.remove(kind)
-            portmap = db.grid[dbrow][dbcol].bels[bel].portmap
+            if kind == 'MIPI_IBUF':
+                portmap = {}
+                portmap['I'] = db.grid[dbrow][dbcol].bels["IOBA"].portmap['I']
+                portmap['IB'] = db.grid[dbrow][dbcol].bels["IOBB"].portmap['I']
+                portmap['OEN'] = db.grid[dbrow][dbcol].bels["IOBA"].portmap['OE']
+                portmap['OENB'] = db.grid[dbrow][dbcol].bels["IOBB"].portmap['OE']
+                portmap['OH'] = db.grid[dbrow][dbcol].bels["IOBA"].portmap['O']
+                portmap['OB'] = db.grid[dbrow][dbcol].bels["IOBB"].portmap['O']
+                portmap['HSREN'] = db.grid[dbrow][dbcol].bels["IOLOGICB"].portmap['SETN']
+                portmap['OL'] = db.grid[dbrow][dbcol + 1].bels["IOBA"].portmap['O']
+            elif kind == 'MIPI_OBUF':
+                portmap = {}
+                portmap['I'] = db.grid[dbrow][dbcol].bels["IOBA"].portmap['I']
+                portmap['IB'] = db.grid[dbrow][dbcol].bels["IOBB"].portmap['I']
+                portmap['IL'] = db.grid[dbrow][dbcol].bels["IOBA"].portmap['I']
+                portmap['MODESEL'] = db.grid[dbrow][dbcol].bels["IOBA"].portmap['OE']
+            elif kind == 'I3C_IOBUF':
+                portmap = db.grid[dbrow][dbcol].bels[bel].portmap.copy()
+                portmap['MODESEL'] = portmap['OE']
+                portmap.pop('OE', None)
+            else:
+                portmap = db.grid[dbrow][dbcol].bels[bel].portmap
             name = f"R{row}C{col}_{kind}_{idx}"
             wires = set(iobmap[kind]['wires'])
             ports = set(chain.from_iterable(iobmap[kind].values())) - wires
@@ -957,7 +1137,20 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
                 wires_suffix = ''
             for port in wires:
                 wname = portmap[port]
-                iob.portmap[portname(port)] = f"R{row}C{col}_{wname}{wires_suffix}"
+                if kind == 'MIPI_IBUF':
+                    if port == 'OH':
+                        wires_suffix_mipi = wires_suffix
+                    else:
+                        wires_suffix_mipi = ''
+                    iob.portmap[portname(port)] = f"R{row}C{col}_{wname}{wires_suffix_mipi}"
+                elif kind == 'MIPI_OBUF':
+                    if port == 'I':
+                        wires_suffix_mipi = wires_suffix
+                    else:
+                        wires_suffix_mipi = ''
+                    iob.portmap[portname(port)] = f"R{row}C{col}_{wname}{wires_suffix_mipi}"
+                else:
+                    iob.portmap[portname(port)] = f"R{row}C{col}_{wname}{wires_suffix}"
 
             for port in ports:
                 iob.portmap[port] = f"R{row}C{col}_{port}{idx}"
@@ -970,7 +1163,8 @@ def tile2verilog(dbrow, dbcol, bels, pips, clock_pips, mod, cst, db):
             mod.primitives[name] = iob
             # constraints
             pos = chipdb.loc2pin_name(db, dbrow, dbcol)
-            bank = chipdb.loc2bank(db, dbrow, dbcol)
+            # XXX tangnano4k uses IOT30 not found in the package
+            #bank = chipdb.loc2bank(db, dbrow, dbcol)
             cst.ports[name] = f"{pos}{idx}"
             if kind[0:5] == 'TLVDS':
                 cst.ports[name] = f"{pos}{idx},{pos}{chr(ord(idx) + 1)}"
@@ -1029,12 +1223,20 @@ def fix_plls(db, mod):
         fix_pll_ports(pll)
 
 def main():
+    pil_available = True
+    try:
+        from PIL import Image
+    except ImportError:
+        pil_available = False
+
     parser = argparse.ArgumentParser(description='Unpack Gowin bitstream')
     parser.add_argument('bitstream')
     parser.add_argument('-d', '--device', required=True)
     parser.add_argument('-o', '--output', default='unpack.v')
     parser.add_argument('-s', '--cst', default=None)
     parser.add_argument('--noalu', action = 'store_true')
+    if pil_available:
+        parser.add_argument('--png')
 
     args = parser.parse_args()
 
@@ -1047,28 +1249,64 @@ def main():
         luts = m.group(3)
         _device = f"GW1N{mods}-{luts}"
 
-    with importlib.resources.open_binary("apycula", f"{_device}.pickle") as f:
-        db = pickle.load(f)
+    with importlib.resources.path('apycula', f'{args.device}.pickle') as path:
+        with closing(gzip.open(path, 'rb')) as f:
+            db = pickle.load(f)
 
     global _pinout
     _pinout = db.pinout[_device][_packages[_device]]
 
-    bitmap = read_bitstream(args.bitstream)[0]
+    bitmap, _, _, extra_slots = read_bitstream(args.bitstream)
     bm = chipdb.tile_bitmap(db, bitmap)
     mod = codegen.Module()
     cst = codegen.Constraints()
 
+    if pil_available and args.png:
+        display(args.png, bitmap)
+
+    # make wire aliases from Himbaechel nodes
+    def by_name_len(el):
+        return len(el[2])
+
+    for node_desc in db.nodes.values():
+        root_wire = None
+        for row, col, wire in sorted(node_desc[1], key = by_name_len):
+            wire_name = f'R{row + 1}C{col + 1}_{wire}'
+            if not root_wire:
+                root_wire = wire_name
+                continue
+            mod.wire_aliases[wire_name] = root_wire
+    for row in range(db.rows):
+        for col in range(db.cols):
+            for i in [1, 2]:
+                mod.wire_aliases[chipdb.wire2global(row + 0, col + 1, db, f'N1{i}1')] = f'R{row + 1}C{col + 1}_SN{i}0'
+                mod.wire_aliases[chipdb.wire2global(row + 2, col + 1, db, f'S1{i}1')] = f'R{row + 1}C{col + 1}_SN{i}0'
+                mod.wire_aliases[chipdb.wire2global(row + 1, col + 0, db, f'W1{i}1')] = f'R{row + 1}C{col + 1}_EW{i}0'
+                mod.wire_aliases[chipdb.wire2global(row + 1, col + 2, db, f'E1{i}1')] = f'R{row + 1}C{col + 1}_EW{i}0'
+
+    # Slots have no wires only func fuses
+    if extra_slots:
+        for slot_idx, slot_bitmap in extra_slots.items():
+            if slot_idx in {2, 3, 4, 5, 6, 8}:
+                av = parse_attrvals(slot_bitmap, db.rev_logicinfo('PLL'), db.shortval[1024]['PLL'], attrids.pll_attrids, "PLL")
+                print('Slot:', slot_idx, av)
+            elif slot_idx in {1}:
+                av = parse_attrvals(slot_bitmap, db.rev_logicinfo('ADC'), db.shortval[1026]['ADC'], attrids.adc_attrids, "ADC")
+                print('Slot:', slot_idx, av)
+            else:
+                print('Unknown Slot:', slot_idx)
+
     # XXX this PLLs have empty main cell
     if _device in {'GW1N-9C', 'GW1N-9'}:
         bm_pll = chipdb.tile_bitmap(db, bitmap, empty = True)
         bm[(9, 0)] = bm_pll[(9, 0)]
         bm[(9, 46)] = bm_pll[(9, 46)]
-
-    for (drow, dcol, dname), (srow, scol, sname) in db.aliases.items():
-        src = f"R{srow+1}C{scol+1}_{sname}"
-        dest = f"R{drow+1}C{dcol+1}_{dname}"
-        mod.wires.update({src, dest})
-        mod.assigns.append((dest, src))
+    if _device in {'GW2A-18', 'GW2A-18C'}:
+        bm_pll = chipdb.tile_bitmap(db, bitmap, empty = True)
+        bm[(9, 0)] = bm_pll[(9, 0)]
+        bm[(9, 55)] = bm_pll[(9, 55)]
+        bm[(45, 0)] = bm_pll[(45, 0)]
+        bm[(45, 55)] = bm_pll[(45, 55)]
 
     # banks first: need to know iostandards
     for pos in db.corners.keys():
@@ -1077,7 +1315,7 @@ def main():
             t = bm[(row, col)]
         except KeyError:
             continue
-        bels, pips, clock_pips = parse_tile_(db, row, col, t)
+        bels, pips, clock_pips = parse_tile_(db, row, col, t, bm)
         #print("bels:", bels)
         tile2verilog(row, col, bels, pips, clock_pips, mod, cst, db)
 
@@ -1086,13 +1324,7 @@ def main():
         # skip banks & dual pisn
         if (row, col) in db.corners:
             continue
-        #for bitrow in t:
-        #    print(*bitrow, sep='')
-        #if idx == (5, 0):
-        #    from fuse_h4x import *
-        #    fse = readFse(open("/home/pepijn/bin/gowin/IDE/share/device/GW1N-1/GW1N-1.fse", 'rb'))
-        #    breakpoint()
-        bels, pips, clock_pips = parse_tile_(db, row, col, t, noiostd = False)
+        bels, pips, clock_pips = parse_tile_(db, row, col, t, bm, noiostd = False)
         #print("bels:", idx, bels)
         #print(pips)
         #print(clock_pips)
